/*
 * =====================================================================================
 *
 *       Filename:  vector.cpp
 *
 *    Description:  Triangle rasterization (filled rotating cube)
 *
 *        Version:  1.0
 *        Created:  23.12.2016 13:12:47
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include <stddef.h>
#include <stdio.h>
extern "C" {
#include <v9990.h>
#include "sintab64.h"
}

VDP_BOX screen;

typedef struct VECTOR
{
    int16_t x;
    int16_t y;
    int16_t z;
} VECTOR;

typedef struct SURFACE
{
    uint8_t vertice[3];
    VECTOR normal;
} SURFACE;

VECTOR vertices[8] =
{
    { -50,  50,  50 },
    {  50,  50,  50 },
    {  50, -50,  50 },
    { -50, -50,  50 },
    { -50,  50, -50 },
    {  50,  50, -50 },
    {  50, -50, -50 },
    { -50, -50, -50 },
};

VECTOR vertices2[8];

/* Define planes.
   { x, y, z, normal-x, normal-y, normal-z } */
SURFACE cube[12] =
{
    { 3, 1, 0, 0, 0, 0 },
    { 2, 1, 3, 0, 0, 0 },
    { 5, 4, 0, 0, 0, 0 },
    { 1, 5, 0, 0, 0, 0 },
    { 6, 3, 7, 0, 0, 0 },
    { 2, 3, 6, 0, 0, 0 },
    { 7, 4, 5, 0, 0, 0 },
    { 5, 6, 7, 0, 0, 0 },
    { 0, 4, 7, 0, 0, 0 },
    { 3, 0, 7, 0, 0, 0 },
    { 5, 1, 2, 0, 0, 0 },
    { 6, 5, 2, 0, 0, 0 },
};

typedef struct MOVEMENT
{
    uint16_t duration;
    int8_t dx;
    int8_t dy;
    int8_t dz;
    int8_t tz;
} MOVEMENT;

/* Just arbitrary values for movements.
   { duration[frames], x-rotation, y-rotation, z-rotation, z-translation } */
MOVEMENT movement[] =
{
    {  500,  12,  12, -11,  0 },
    {  200,  14,  18,  13,  0 },
    {  200,  12,  10,  04,  0 },
    {  200,  20,  00, -10,  0 },
    {  200,  10, -13, -23,  0 },
    {  200,  20, -23, -17,  0 },
    {  200,  10, -27, -15,  0 },
    {  200,  10,  00, -20,  0 },
    {  200,  10,  10, -30,  0 },
    {  200,  20,  30, -20,  0 },
    {  200,  10,  20, -10,  0 },
    {  200,  30,  20, -20,  0 },
    {  200,  20,  20, -30,  0 },
    {  200,  10,  12, -30,  0 },
    {  200,  00,  20, -30,  0 },
    {  200, -10,  20, -30,  0 },
    { 1000, -20,  10,  20,  0 },
    { 1000, -30, -20,  30,  0 },
    { 1000, -30, -10,  20,  0 },
    { 1000, -20,  10,  10,  0 },
    {    0,  0,  0,  0,  0 },
};

uint8_t page=1;
uint8_t volatile flip=0;
uint16_t framecount=0;
uint16_t nextframe=0;

inline void flip_vidpages()
{
    wait_vdp();

    if (page == 0)
    {
        VDPWriteReg (VDP_SCROLL_HIGH_Y,1);
        page=1;
    }
    else
    {
        VDPWriteReg (VDP_SCROLL_HIGH_Y,2);
        page=0;
    }
}

void ISR1 (void) __attribute__ ((interrupt));
void ISR1 (void)
{
    if (flip)
    {
        flip_vidpages();
        flip=0;
    }

    framecount++;
    WRITE_VDP (VDP_IFP,0x07);	// IRQ loeschen
}

inline uint8_t SearchBordercolor (uint8_t x, uint8_t y, uint8_t direction)
{
    wait_vdp();
    VDPWriteReg (VDP_SC_X, (uint8_t)x);
    VDPWriteReg (VDP_SC_Y, (uint8_t)y);
    VDPWriteReg (VDP_ARG, (uint8_t)direction);
    VDPWriteReg (VDP_FC, (uint8_t)0);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    VDPWriteReg (VDP_OPCODE,VDP_OPCODE_SRCH);
    wait_vdp();

    if (! (READ_VDP (VDP_STATUS) & VDP_STATUS_BD))
    {
        /*  Border color not detected */
        return -1;
    }

    return VDPReadReg (VDP_BDC_LOW_X);
}

inline void DrawLine (uint8_t x1, uint8_t y1, short x2, short y2, short color)
{
    short mj, mi;
    uint8_t maj=0;

    if (x2<0)
    {
        maj += 4;
        x2 = -1*x2;
    }

    if (y2<0)
    {
        maj +=8;
        y2 = -1*y2;
    }

    if (x2>=y2)
    {
        mj = x2;
        mi = y2;
        maj += 0;
    }
    else
    {
        mj = y2;
        mi = x2;
        maj += 1;
    }

    wait_vdp();
    VDPWriteReg (VDP_DS_X, (uint8_t)x1);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)y1);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)mj);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)mi);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)maj);
    VDPWriteReg (VDP_FC, (uint8_t)color);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (color>>8));
    VDPWriteReg (VDP_OPCODE,VDP_OPCODE_LINE);
    wait_vdp();
}

inline void fillline (uint8_t x1, uint8_t y1, uint8_t length, short color)
{
    wait_vdp();
    VDPWriteReg (VDP_DS_X, (uint8_t)x1);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)y1);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)length);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)0);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t)0);
    VDPWriteReg (VDP_FC, (uint8_t)color);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (color>>8));
    VDPWriteReg (VDP_OPCODE,VDP_OPCODE_LINE);
}

void DrawTriangleScratchpad (VECTOR p0, VECTOR p1, VECTOR p2, short color)
{
    static uint16_t y, page_y;
    uint8_t i, y1, y2, y3, x1, x2;
    static VDP_COPY_XY_XY bltBox;
    page_y=256;

    if (page==1)
    {
        page_y=512;
    }

    /* Draw triangle to scratch pad */
    DrawLine (p0.x, p0.y, p1.x-p0.x, p1.y-p0.y, color);
    DrawLine (p1.x, p1.y, p2.x-p1.x, p2.y-p1.y, color);
    DrawLine (p2.x, p2.y, p0.x-p2.x, p0.y-p2.y, color);
    /* Fill triangle. Improvement would be to compute the
       complete bounding box.
       - Iterate from min-y to max-y
       - Get first x-value searching from right to left
       - Get second x-value searching from left to right
       - Draw a line between both x-values */
    y1 = p0.y;
    y2 = p0.y;

    if (p0.y >= p1.y)
    {
        y1 = p1.y;
    }

    if (y1 >= p2.y)
    {
        y1 = p2.y;
    }

    if (y2 < p1.y)
    {
        y2 = p1.y;
    }

    if (y2 < p2.y)
    {
        y2 = p2.y;
    }

    VDPWriteReg (VDP_SC_X, (uint8_t)0);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    VDPWriteReg (VDP_NY, (uint8_t)x1);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (0));
    VDPWriteReg (VDP_FC, (uint8_t)color);
    WRITE_VDP (VDP_REGDAT, (uint8_t) (color>>8));

    for (i=y1; i<y2; i++)
    {
        x2 = SearchBordercolor (240, i, 6);
        x1 = SearchBordercolor (10, i, 2);
        fillline (x1, i, x2-x1, color);
    }

    /* Copy scratchpad to buffer */
    v9990_SetCmdWriteMask (0xffff);
    VDPWriteReg (VDP_ARG,0);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCORSC);
    bltBox.sourceX=70;
    bltBox.sourceY=y1;
    bltBox.width=150;
    bltBox.height=y2-y1+1;
    bltBox.destX=70;
    bltBox.destY=y1+page_y;
    v9990_CopyXYToXY (&bltBox);
    wait_vdp();
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    /* clear scratchpad. Improvement: Work on bounding box area only. */
    screen.top=0;
    screen.left=0;
    v9990_DrawFilledBox (&screen,0);
}


void clear_page (uint8_t page)
{
    screen.left=0;

    if (page==0)
    {
        screen.top=255;
    }
    else
    {
        screen.top=512;
    }

    v9990_DrawFilledBox (&screen,0);
}

/* Compute vector normals */
void normalvector (SURFACE *flaeche)
{
    short x, y, z, x2, y2, z2;
    x=vertices[flaeche->vertice[1]].x-vertices[flaeche->vertice[0]].x;
    y=vertices[flaeche->vertice[1]].y-vertices[flaeche->vertice[0]].y;
    z=vertices[flaeche->vertice[1]].z-vertices[flaeche->vertice[0]].z;
    x2=vertices[flaeche->vertice[2]].x-vertices[flaeche->vertice[1]].x;
    y2=vertices[flaeche->vertice[2]].y-vertices[flaeche->vertice[1]].y;
    z2=vertices[flaeche->vertice[2]].z-vertices[flaeche->vertice[1]].z;
    flaeche->normal.x=y*z2-z*y2;
    flaeche->normal.y=z*x2-x*z2;
    flaeche->normal.z=x*y2-y*x2;
}

int main (void)
{
    unsigned char *buf;
    uint16_t i,angle_x,angle_y,angle_z,j,k,page_y;
    short color, zstep=0;
    int16_t x,y,z,x2,y2,z2;
    int16_t dx, dy, dz, tz;
    uint8_t step=0, updown=0;
    uint8_t trans_x, trans_y, trans_z, seiten;
    trans_x = 100;
    trans_y = 100;
    trans_z = 100;
    dx = 4;
    dy = 0;
    dz = 4;
    tz = 200;
    VDP_COLOR farben[64];

    for (short i=0; i<64; i++)
    {
        farben[i].red=i;
        farben[i].green=i;
        farben[i].blue=i;
    }

    v9990_SetBackdropColor (0);
    v9990_WritePalette (farben, 0, 64);
    v9990_SpritesDisable();
    v9990_SetScreenMode (VDP_MODE_B1, VDP_SCR0_8BIT, VDP_SCR0_XIM512, VDP_SCR1_PAL, 0);
    asm volatile ("move.w   #0x2700, %sr"); // Interrupts off
    (* ((long *)0x64) = (long)ISR1);
    VDPWriteReg (VDP_INT_ENABLE,VDP_INT_IEV);
    WRITE_VDP (VDP_IFP,0x07);
    asm volatile ("move.w   #0x2000, %sr"); // Interrupts on
    screen.left=0;
    screen.width=256;
    screen.height=240;
    clear_page (0);
    clear_page (1);
    angle_x=0,angle_y=0,angle_z=0;

    while (1)
    {
        clear_page (page);
        // z-Rotate vertices
        j=angle_z&511;
        k= (angle_z+128)&511;

        for (i=0; i<8; i++) /* Ecken drehen */
        {
            vertices2[i].x= (vertices[i].x*aSin[k]-vertices[i].y*aSin[j])>>6;
            vertices2[i].y= (vertices[i].x*aSin[j]+vertices[i].y*aSin[k])>>6;
            vertices2[i].z=vertices[i].z;
        }

        // x-Rotate vertices
        j=angle_x&511;
        k= (angle_x+128)&511;

        for (i=0; i<8; i++)
        {
            vertices2[i].x=vertices2[i].x;
            y=vertices2[i].y;
            vertices2[i].y= (vertices2[i].y*aSin[k]-vertices2[i].z*aSin[j])>>6;
            vertices2[i].z= (y*aSin[j]+vertices2[i].z*aSin[k])>>6;
        }

        // y-Rotate vertices
        j=angle_y&511;
        k= (angle_y+128)&511;

        for (i=0; i<8; i++)
        {
            x=vertices2[i].x;
            vertices2[i].x= (vertices2[i].x*aSin[k]-vertices2[i].z*aSin[j])>>6;
            vertices2[i].y=vertices2[i].y;
            vertices2[i].z= (x*aSin[j]+vertices2[i].z*aSin[k])>>6;
        }

        //Translation (Z-Axis)
        for (seiten=0; seiten<8; seiten++)
        {
            vertices2[seiten].z += tz;
        }

        //3D->2D Transformation
        for (i=0; i<8; i++)
        {
            vertices2[i].x= (vertices2[i].x<<7)/ (vertices2[i].z+150)+128;
            vertices2[i].y= (vertices2[i].y<<7)/ (vertices2[i].z+150)+121;
        }

        // Calculate vector normals
        for (seiten=0; seiten<12; seiten++)
        {
            x=vertices2[cube[seiten].vertice[0]].x-vertices2[cube[seiten].vertice[1]].x;
            y=vertices2[cube[seiten].vertice[0]].y-vertices2[cube[seiten].vertice[1]].y;
            x2=vertices2[cube[seiten].vertice[2]].x-vertices2[cube[seiten].vertice[1]].x;
            y2=vertices2[cube[seiten].vertice[2]].y-vertices2[cube[seiten].vertice[1]].y;
            cube[seiten].normal.z=x*y2-y*x2;
        }

        angle_x+=dx;
        angle_x &=511;
        angle_y+=dy;
        angle_y &=511;
        angle_z+=dz;
        angle_z &=511;
        v9990_SetCmdWriteMask (0xffff);
        VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
        page_y=0;

        /* Render Triangle */
        for (k=0; k<12; k++)
        {
            if (cube[k].normal.z>0)
            {
                color  = 2 + (k>>1)%2;
                color |= ((k>>1)%5) + (color << 3);
                color |= ((k>>1)%5) + (color << 2);
                color |= (color << 8);
                DrawTriangleScratchpad (vertices2[cube[k].vertice[0]], vertices2[cube[k].vertice[1]], vertices2[cube[k].vertice[2]], color);
            }
        }

        wait_vdp();
        flip=1;

        if (framecount > nextframe)
        {
            if (movement[step].duration == 0)
            {
                step = 0;
                framecount = 0;
                nextframe = 0;
            }

            nextframe = movement[step].duration + framecount;
            dx = movement[step].dx;
            dy = movement[step].dy;
            dz = movement[step].dz;
            //tz = movement[step].tz;
            step++;
        }

        zstep += 20;
        zstep &= 511;
        tz = 200 + (aSin[zstep++] << 1);

        /* Here you could make use of a triple buffer.
           Start to render the next frame to a third
           buffer, instead of waiting until the buffers
           have been swapped. */
        while (flip);
    }

    return 0;
}
